@@ -684,26 +684,6 @@ int ldb_next_del_trans(struct ldb_module *module)
 	return ret;
 }
 
-struct ldb_handle *ldb_handle_new(TALLOC_CTX *mem_ctx, struct ldb_context *ldb)
-{
-	struct ldb_handle *h;
-
-	h = talloc_zero(mem_ctx, struct ldb_handle);
-	if (h == NULL) {
-		ldb_set_errstring(ldb, "Out of Memory");
-		return NULL;
-	}
-
-	h->status = LDB_SUCCESS;
-	h->state = LDB_ASYNC_INIT;
-	h->ldb = ldb;
-	h->flags = 0;
-	h->location = NULL;
-	h->parent = NULL;
-
-	return h;
-}
-
 /* calls the request callback to send an entry
  *
  * params:
@@ -901,7 +881,10 @@ static int ldb_modules_load_path(const char *path, const char *version)
 	} *loaded;
 	struct loaded *le;
 	int dlopen_flags;
+
+#ifdef RTLD_DEEPBIND
 	bool deepbind_enabled = (getenv("LDB_MODULES_DISABLE_DEEPBIND") == NULL);
+#endif
 
 	ret = stat(path, &st);
 	if (ret != 0) {
@@ -937,13 +920,13 @@ static int ldb_modules_load_path(const char *path, const char *version)
 #ifdef RTLD_DEEPBIND
 	/*
 	 * use deepbind if possible, to avoid issues with different
-	 * system library varients, for example ldb modules may be linked
+	 * system library variants, for example ldb modules may be linked
 	 * against Heimdal while the application may use MIT kerberos.
 	 *
 	 * See the dlopen manpage for details.
 	 *
 	 * One typical user is the bind_dlz module of Samba,
-	 * but symbol versioniong might be enough...
+	 * but symbol versioning might be enough...
 	 *
 	 * We need a way to disable this in order to allow the
 	 * ldb_*ldap modules to work with a preloaded socket wrapper.
